#include <stdio.h>
#include <assert.h>
#include <io.h>
#include <fcntl.h>
#include <wtypes.h>
#include <winbase.h>

class CONIN {
public:
   CONIN();
   virtual ~CONIN();
   char *Gets(char *buf);
   int Poll(int timeout);
private:
   HANDLE                   h_;
   HANDLE                   hthread_;
   DWORD                    thread_id_;
   HANDLE                   hevent_;
   BOOL                     ret_;
   DWORD                    cnt_;
   char                     buf_[256];
   friend DWORD WINAPI ThreadConin(LPVOID pconin);
};

CONIN::Poll(
   int                      timeout
) {
   int                      ret = 0;

   switch(WaitForSingleObject(hevent_, timeout/1000)) {
      case WAIT_ABANDONED: {
         break;
      }
      case WAIT_OBJECT_0: {
         ret = cnt_;
         break;
      }
      case WAIT_TIMEOUT: {
         break;
      }
      case WAIT_FAILED: {
         printf("Waitfor WAIT_FAILED %d\n", GetLastError());
         break;
      }
      default: {
         assert(0);
         break;
      }
   }
   return(ret);
}

char *
CONIN::Gets(
   char                    *buf
) {
   if (cnt_) {
      memcpy(buf, buf_, cnt_);
      buf[cnt_ - 2] = 0;
      SetEvent(hevent_);
      return(buf);
   }
   return(0);
}

CONIN::CONIN(
) {
   hevent_ = CreateEvent(
    (LPSECURITY_ATTRIBUTES) 0    
   ,(BOOL)                  FALSE  // Manual-reset.
   ,(BOOL)                  FALSE  // Initial state.
   ,(LPCTSTR)               0      // Name.
   );
   assert(0 != hevent_);

   hthread_ = CreateThread(
    (LPSECURITY_ATTRIBUTES)   0
   ,(DWORD)                   0
   ,(LPTHREAD_START_ROUTINE)  ThreadConin
   ,(LPVOID)                  this
   ,(DWORD)                   0
   ,(LPDWORD)                &thread_id_
   );
}

CONIN::~CONIN(
) {
    if (WaitForSingleObject(hthread_, 5000) == WAIT_TIMEOUT) {
        TerminateThread(hthread_, 1);
   }
}

DWORD WINAPI 
ThreadConin(
   LPVOID                     p
) {
   CONIN                     *pconin;

   pconin = (CONIN *)p;

   pconin->h_ = CreateFile(
    (LPCTSTR)
      "CONIN$"
   ,(DWORD) 0
     | GENERIC_READ
   ,(DWORD) 0
     | FILE_SHARE_READ
     | FILE_SHARE_WRITE
   ,(LPSECURITY_ATTRIBUTES) 0
   ,(DWORD) 0
     | OPEN_EXISTING
   ,(DWORD) 0                    // Flags.
   ,(HANDLE) 0                   // Attributes.
   );
   assert(0 != pconin->h_);

   while (1) {
      pconin->cnt_ = 0;
      memset(pconin->buf_, 0, sizeof(pconin->buf_));
      pconin->ret_ = ReadFile(
       pconin->h_, pconin->buf_, 255, &pconin->cnt_, (LPOVERLAPPED) 0);
      SetEvent(pconin->hevent_);
      WaitForSingleObject(pconin->hevent_, INFINITE);
   }
   return(1);
}

static CONIN conin;

extern "C" {
int
ConinPoll(
   int                      timeout
) {
   return(conin.Poll(timeout));
}

char *
ConinGets(
   char                    *buf
) {
   return(conin.Gets(buf));
}

} //extern "C"

