/*
    Copyright (c) 1997 Eliot W. Dudley.  All rights reserved.

       edudley@servtech.com
       XIDAR
       3388 STATE RT 370
       CATO NY  13033

*/
/*
       1-Wire, DS, DS1820, DS2405, DS2407, DS9097, and MicroLan
          are trademarks and/or registered trademarks of
                 Dallas Semiconductor Coporation
*/
  #include <stdio.h>
  #include <string.h>
  #include <stdarg.h>
  #include <stdlib.h>
  #include <setjmp.h>

  #include <wtypes.h>
  #include <winbase.h>

  #include "msglue.h"
  #include "util.h"
  #include "b.h"

  #include "mlfnc.h"
  #include "ml.h"
  #include "mlerr.h"
  #include "mld.h"
  #include "mlbus.h"

#define dcbpr(dcbmem) printf("%04x " #dcbmem "\n", dcbmem)
void
DCBPr(
   DCB            *pdcb
) {
   dcbpr(pdcb->DCBlength);
   dcbpr(pdcb->BaudRate);
   dcbpr(pdcb->fBinary);
   dcbpr(pdcb->fParity);
   dcbpr(pdcb->fOutxCtsFlow);
   dcbpr(pdcb->fOutxDsrFlow);
   dcbpr(pdcb->fDtrControl);
   dcbpr(pdcb->fDsrSensitivity);
   dcbpr(pdcb->fTXContinueOnXoff);

   dcbpr(pdcb->fOutX);
   dcbpr(pdcb->fInX);
   dcbpr(pdcb->fErrorChar);
   dcbpr(pdcb->fNull);
   dcbpr(pdcb->fRtsControl);
   dcbpr(pdcb->fAbortOnError);
   dcbpr(pdcb->fDummy2);
   dcbpr(pdcb->wReserved);
   dcbpr(pdcb->XonLim);

   dcbpr(pdcb->XoffLim);
   dcbpr(pdcb->ByteSize);
   dcbpr(pdcb->Parity);
   dcbpr(pdcb->StopBits);
   dcbpr(pdcb->XonChar);
   dcbpr(pdcb->XoffChar);
   dcbpr(pdcb->ErrorChar);
   dcbpr(pdcb->EofChar);
   dcbpr(pdcb->EvtChar);

   dcbpr(pdcb->wReserved1);
}

struct MLPORTt {
    /// Serial port device name.
   char                   *name;
    /// Its handle.
   HANDLE                   h;
    /// Per character, per packet timing characterization.
   COMMTIMEOUTS             commtimeouts;
    /// 18N and so forth.
   DCB                      dcb;
    /// Buf for expanding i/o into one 115.2Kbps byte per bit of information.
   BYTE                    buf[8 * ML_MAX_IO];
};

/*/
   Open the serial device.
*/
int
MLOpen(
   pML                    pml,
   char                  *dev
) {
   pMLPORT                pmlport;
   DCB                   *pdcb;
   STACK_CHECK();
   ENTER(MLOpen);

   if (!dev) {
      FATAL(NO_SERIAL_DEVICE_SPECIFIED, "");
   }
   CALLOC(pmlport, 1, MLPORT);
   pml->pmlport = pmlport;
   MALLOC(pmlport->name, strlen(dev) + 1, char);
   strcpy(pmlport->name, dev);

   pmlport->h = CreateFile(
    pmlport->name
   ,GENERIC_READ | GENERIC_WRITE
   ,0
   ,NULL
   ,OPEN_EXISTING
   ,FILE_ATTRIBUTE_NORMAL
   ,0
   );

   pmlport->commtimeouts.ReadIntervalTimeout = 20;
   pmlport->commtimeouts.ReadTotalTimeoutMultiplier = 1;
   pmlport->commtimeouts.ReadTotalTimeoutConstant =  10;
   pmlport->commtimeouts.WriteTotalTimeoutMultiplier = 0;
   pmlport->commtimeouts.WriteTotalTimeoutConstant = 0;
   SetCommTimeouts(pmlport->h, &pmlport->commtimeouts);

   pdcb = &pmlport->dcb;
   GetCommState(pmlport->h, pdcb);
   pdcb->DCBlength = sizeof(DCB);               // sizeof(DCB)
   pdcb->BaudRate = CBR_115200;                 // current baud rate
   pdcb->fBinary = 1;                           // binary mode, no EOF check
   pdcb->fParity = 0;                           // enable parity checking
   pdcb->fOutxCtsFlow = 0;                      // CTS output flow control
   pdcb->fOutxDsrFlow = 0;                      // DSR output flow control
   pdcb->fDtrControl = DTR_CONTROL_ENABLE;      // DTR flow control type
   pdcb->fDsrSensitivity = 0;                   // DSR sensitivity
   pdcb->fTXContinueOnXoff = 0;                 // XOFF continues Tx
   pdcb->fOutX = 0;                             // XON/XOFF out flow control
   pdcb->fInX = 0;                              // XON/XOFF in flow control
   pdcb->fErrorChar = 0;                        // enable error replacement
   pdcb->fNull = 0;                             // enable null stripping
   pdcb->fRtsControl = RTS_CONTROL_ENABLE;      // RTS flow control
   pdcb->fAbortOnError = 0;                     // abort reads/writes on error
   pdcb->fDummy2 = 0;                           // reserved
   pdcb->wReserved = 0;                         // not currently used
   pdcb->XonLim =  0;                           // transmit XON threshold
   pdcb->XoffLim =  0;                          // transmit XOFF threshold
   pdcb->ByteSize = 6;                          // number of bits/byte, 4-8
   pdcb->Parity = NOPARITY;                     // 0-4=no,odd,even,mark,space
   pdcb->StopBits = ONESTOPBIT;                 // 0,1,2 = 1, 1.5, 2
   pdcb->XonChar = 0;                           // Tx and Rx XON character
   pdcb->XoffChar = 0;                          // Tx and Rx XOFF character
   pdcb->ErrorChar = 0;                         // error replacement character
   pdcb->EofChar = 0;                           // end of input character
   pdcb->EvtChar = 0;                           // received event character
   pdcb->wReserved1 = 0;                        // reserved; do not use
   if (!(SetCommState(pmlport->h, pdcb))) {
      ;
   }

   PurgeComm(pmlport->h, PURGE_TXCLEAR | PURGE_RXCLEAR);

   POP_CALL_TRACE();
   return(1);
}

//</XMP><A NAME="MLWrite"><XMP>
/*/
   Write count bits from buf to the bus object pml.  Read the state of
    the bus for each bit written and record in the same buf.  According
    to flag options, optionally reset the bus before the write, and/or
    timestamp the pmlnode passed by reference.  Returns error code less
    than zero on error, else the number of bits written.
    Source: \URL{../mld.c.html#MLWrite}
*/
int
MLWrite(
   pML                     pml,
   char                   *buf,
   int                     count,
   int                     flag,
   pMLNODE                 pmlnode
) {
   int                     ret;
   int                     i;
   int                     cnt_wr;
   int                     cnt_rd;
   BYTE                    w;
   BYTE                    r;
   struct timeval          tv;
   pMLPORT                 pmlport;
   DCB                    *pdcb;

   pmlport = pml->pmlport;
   pdcb = &pmlport->dcb;

   //==========================================================================
   if ((ML_RESET_ADDR & flag) || (ML_RESET_POWER & flag)) {
      //=Go to eight bit. 10.47KBaud for reset addr, 450 for reset power.======
      PurgeComm(pmlport->h, PURGE_TXCLEAR | PURGE_RXCLEAR);
      GetCommState(pmlport->h, pdcb);
      pdcb->BaudRate = (ML_RESET_ADDR & flag) ? 9600 : CBR_300;
      pdcb->ByteSize = 8; 
      if (!(SetCommState(pmlport->h, pdcb))) {
         ;
      }
      //=======================================================================

      //=Send the reset pulse.=================================================
      // (115200Hz clock)/(Divisor = 11) --> 95.5uSec per bit.
      // LSB first, start bit is low, stop bit is high.
      // So five bits low, five bits high is 477uSec reset pulse.

      // (115200Hz clock)/(Divisor = 256) --> 2222.2 uSec per bit.
      // LSB first, start bit is low, stop bit is high.
      // So nine bits low, one bit high is 20ms power reset pulse.

      w = (ML_RESET_ADDR & flag) ? 0xf0 : 0x00;
      ret = WriteFile(pmlport->h, &w, 1,  &cnt_wr, NULL);
      // Convert bits to byte stream while waiting.
      for (i = 0; i < count; ++i) {
         pmlport->buf[i] = "\x00\xff"[test_bit(i, buf)];
      }
      ret = ReadFile(pmlport->h, &r, 1,  &cnt_rd, NULL);

      // We're not getting echo on the bus?
      if (!cnt_rd) {
         return(-XIDARML_BUS_FAILURE);
      }
      if (ML_RESET_ADDR & flag) {
         // No reset pulse detected, we're talking to the bus but no answer.
         if (r == w) {
            return(-NO_DEVICES_FOUND);
         }
      } else {
         return(1);
      }
      GetCommState(pmlport->h, pdcb);
      pdcb->BaudRate = CBR_115200;
      pdcb->ByteSize = 6; 
      if (!(SetCommState(pmlport->h, pdcb))) {
         ;
      }
   } else {
      // Convert bits to byte stream for write.
      for (i = 0; i < count; ++i) {
         pmlport->buf[i] = "\x00\xff"[test_bit(i, buf)];
      }
   }

   ret = WriteFile(pmlport->h, pmlport->buf, count,  &cnt_wr, NULL);
   // Wait for response.  Read them back.
   memset(buf, 0, (count + 7) / 8);
   if ((pmlnode) && (ML_TIMESTAMP & flag)) {
      gettimeofday(&tv, (struct timezone *)0);
      pmlnode->tv = tv;
   }
   ret = ReadFile(pmlport->h, pmlport->buf, count,  &cnt_rd, NULL);
   if ((count != cnt_wr) || (cnt_wr != cnt_rd)) {
      // We're not getting echo on the bus.
      return(-XIDARML_BUS_FAILURE);
   }
   for (i = 0; i < count; ++i) {
      if (b_00000001 & pmlport->buf[i]) {
         set_bit(i, buf);
      }
   }
   return(count);
}
