//=============================================================================
  #include <stdio.h>
  #include <errno.h>
  #include <stdarg.h>
  #include <setjmp.h>
#ifdef LINUX
  #include <unistd.h>
  #include <netdb.h>
  #include <sys/time.h>
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #define INVALID_SOCKET -1
  #define SOCKET_ERROR -1
#else
  #include <wtypes.h>
  #include <winbase.h>
  #include <winsock.h>
#endif

  #include "mlfnc.h"
  #include "ml.h"
  #include "mlerr.h"
//=============================================================================

//=============================================================================
typedef struct hostent HOSTENT, *pHOSTENT;
typedef struct protoent PROTOENT, *pPROTOENT;
typedef struct servent SERVENT, *pSERVENT;
typedef struct sockaddr SOCKADDR, *pSOCKADDR;
typedef struct sockaddr_in SOCKADDR_IN, *pSOCKADDR_IN;
//=============================================================================

//=============================================================================
struct SRCVt {
   SOCKET                 s;
   pSOCKADDR              psockaddr;
   pHOSTENT               phostent;
   pPROTOENT              pprotoent;
   pSERVENT               pservent;
   #ifdef WIN32
   WSADATA                wsadata;
   #endif
};

struct SXMTt {
   pSRCV                  psrcv;
   pSOCKADDR              psockaddr;
   int                    sockaddr_len;
};
//=============================================================================

//=============================================================================
pSRCV
SRcvNew(
   void
) {
   pSRCV                  psrcv;

   CALLOC(psrcv, 1, SRCV);
   psrcv->s = -1;
   return(psrcv);
}
pSXMT
SXmtNew(
   void
) {
   pSXMT                  psxmt;

   CALLOC(psxmt, 1, SXMT);
   CALLOC(psxmt->psockaddr, 1, SOCKADDR);
   return(psxmt);
}

pSRCV
SRcvOpen(
   pSRCV                  psrcv,
   char                  *service
) {
   int                    ret;
   char                   buf[256];
   pHOSTENT               phostent;
   pSERVENT               pservent;
   pPROTOENT              pprotoent;
   pSOCKADDR_IN           psockaddr_in;

   #ifdef WIN32
   if (0 != (ret = WSAStartup((2 << 8) | 2, &psrcv->wsadata))) {
      Fatal(ret, "WSAStarup() failed");
   }
   #endif
   if (0 != (ret = gethostname(buf, sizeof(buf)))) {
      Fatal(ret, "gethostname() failed");
   }

   if (0 == (phostent = gethostbyname(buf))) {
      Fatal(errno, "gethostbyname() failed");
   }
   CALLOC(psrcv->phostent, 1, HOSTENT);
   *psrcv->phostent = *phostent;

   if (0 == (pprotoent = getprotobyname("udp"))) {
      Fatal(errno, "getprotobyname() failed");
   }
   CALLOC(psrcv->pprotoent, 1, PROTOENT);
   *psrcv->pprotoent = *pprotoent;

   if (0 == (pservent = getservbyname(service, "udp"))) {
      Fatal(errno, "getservbyname(\"%s\", \"udp\") failed", service);
   }
   CALLOC(psrcv->pservent, 1, SERVENT);
   *psrcv->pservent = *pservent;

   if (INVALID_SOCKET ==
    (psrcv->s = socket(AF_INET, SOCK_DGRAM, psrcv->pprotoent->p_proto))) {
      Fatal(errno, "socket() failed");
   }

   CALLOC(psrcv->psockaddr, 1, SOCKADDR);
   psockaddr_in = (pSOCKADDR_IN)psrcv->psockaddr;
   psockaddr_in->sin_family = AF_INET;
   psockaddr_in->sin_port = psrcv->pservent->s_port;
   psockaddr_in->sin_addr = *((struct in_addr*)*psrcv->phostent->h_addr_list);

   if (SOCKET_ERROR == bind(psrcv->s, psrcv->psockaddr, sizeof(*psrcv->psockaddr))) {
      Fatal(errno, "bind() failed");
   }
   return(psrcv);
}

SOCKET
SRcvGetSocket(
   pSRCV                      psrcv
) {
   return(psrcv->s);
}

char *
SRcv(
   char                  *buf,
   pSRCV                  psrcv,
   pSXMT                  psxmt
) {
   int                    i;
   int                    j;
   int                    ret;

   psxmt->psrcv = psrcv;
   psxmt->sockaddr_len = sizeof(*psxmt->psockaddr);
   for (*buf = 0, j = 0, i = 0; i < BUF_SIZ;) {
//    if (ReadYield(psrcv->s, 0)) {
         if (0 > (ret = recvfrom(psrcv->s, buf + i,
          BUF_SIZ - i, 0, psxmt->psockaddr, &psxmt->sockaddr_len))) {
            FATAL(RECVFROM_FAILED, "");
         }
         i += ret;
         for (; j < i; ++j) {
            if ((buf[j] == '\r') || (buf[j] == '\n')) {
               buf[j] = 0;
               return(buf);
            }
         }
//    }
   }
   return(0);
}

int
SXmt(
   char                  *buf,
   pSXMT                  psxmt
) {
   return(sendto(psxmt->psrcv->s, buf, strlen(buf), 0,
    psxmt->psockaddr, psxmt->sockaddr_len));
}
//=============================================================================

//=============================================================================
/*
int ReadYield(int fd, int ms) {return(1);}
int
main(
   int                    argc,
   char                 **argv
) {
   int                    ret;
   char                   buf[256];
   pSRCV                  psrcv;
   pSXMT                  psxmt;
   fd_set                 fd_set_rd;
   fd_set                *pfd_set_rd;
   struct timeval         tv;
   char                  *ptr;

   pfd_set_rd = &fd_set_rd;

   psrcv = SRcvNew();
   psxmt = SXmtNew();
   SRcvOpen(psrcv, "xidarml");

   for (;;) {
      FD_ZERO(pfd_set_rd);
      FD_SET(psrcv->s, pfd_set_rd);
      tv.tv_sec = 10;
      tv.tv_usec = 0;
      ret = select(psrcv->s + 1, pfd_set_rd, 0, 0, &tv);
      if (0 > ret) {
         puts("select -1");
         return(ret);
      }
      if (0 == ret) {
         puts("select 0");
         continue;
      }
      SRcv(buf, psrcv, psxmt);
      puts(buf);
      SXmt("this is taipei\n", psxmt);
   }
}
*/
//=============================================================================
