//<XMP>
/*
    mltask.c:  XidarML tasks.
  
    Copyright (c) 1997 Eliot W. Dudley.  All rights reserved.

       edudley@servtech.com 
       XIDAR TECHNICS
       3388 STATE RT 370 
       CATO NY  13033
*/  
/*
       1-Wire, DS, DS1820, DS2405, DS2407, DS9097, and MicroLan 
          are trademarks and/or registered trademarks of
                 Dallas Semiconductor Coporation
*/

//=============================================================================
  #include <stdio.h>
  #include <string.h>
  #include <memory.h>
  #include <stdarg.h>
  #include <stdlib.h>
  #include <limits.h>
  #include <setjmp.h>
  #include <errno.h>
  #include <math.h>

  #include <time.h>
#ifdef LINUX
  #include <unistd.h>
  #include <sys/time.h>
  #include <sys/types.h>
  #include <asm/bitops.h>
#else
  #include "msglue.h"
#endif
 
  #include "util.h"

  #include "mlfnc.h"
  #include "ml.h"
  #include "mld.h"
  #include "mlbus.h"
  #include "mlerr.h"
//=============================================================================

//=============================================================================
  #define T1820_SIZ             0x0400
  #define T2407_SIZ             0x0400
//=============================================================================

//=============================================================================
//</XMP><A NAME="TS1820Log"><XMP>
/*/
   A simple demo of a XidarML task object that fires up and logs temperature
    readings from a DS1820.  The original ml.rc command line would have
    been something like:

    spawn t1820log  chickencoop tmlq tmlmsg chickencoop /data/temperature 1800

   Which means start a task object using function T1820Log() and 
    pass it the command:

    chickencoop tmlq tmlmsg chickencoop /data/temperature 1800

   Which to T1820Log means to start a task named 'chickencoop' and talk to
    the task named 'tml' via message queue 'tmlq' and and message mailbox
    'tmlmsg', and get data from the 1-Wire device also named 'chickencoop',
    log the data to the directory '/data/temperature' in file 'chickencoop.
    And log it every half hour.
    Source: \URL{../mltask.c.html#TS1820Log}
*/

int
T1820Log(
   char                         *cmd,
   int                           pri_ret
) {
   pTCB                          volatile ptcb;
   pEVENT                        volatile ptmlq = 0;
   pEVENT                        volatile ptmlmsg = 0;
   int                           volatile delay = 0;
   char                         *volatile file_name = 0;
   char                         *volatile name= 0;
   FILE                         *volatile LOG;

   int                           ret;
   char                         *ptr;
   pML                           pml;
   pMLNODE                       pmlnode;
   struct tm                    *ptm;
   MLREAD                        mlread;

   if(0 == setjmp((ptcb = TaskCreate(&cmd, T1820_SIZ))->jmp_buf)) {
      ptcb->pmsg = EventCreate(EVENT_NAME_DEFAULT, EVENT_MSG);
      SpawnArg(pri_ret, ptcb, ptr = Shift(&cmd), "null daemon queue name");
      SpawnArg(pri_ret, ptcb, ptmlq = GetEventByName(ptr), 
       "daemon queue not found");
      SpawnArg(pri_ret, ptcb, ptr = Shift(&cmd), "null daemon mailbox name");
      SpawnArg(pri_ret, ptcb, ptmlmsg = GetEventByName(ptr), 
       "daemon mailbox not found");
      SpawnArg(pri_ret, ptcb, ptr = Shift(&cmd), "null log file name");
      MALLOC(name, strlen(ptr) + 1, char);
      strcpy(name, ptr);
      SpawnArg(pri_ret, ptcb, ptr = Shift(&cmd), "null log file directory");
      MALLOC(file_name, strlen(ptr) + strlen(name) + 2, char);
      strcat(strcat(strcpy(file_name, ptr), "/"), name);
      delay = (delay = atoi((ptr = Shift(&cmd)) ? ptr : "10")) ? delay : 10;
      SpawnAOK(pri_ret, ptcb);
   }
   StackClr(ptcb);
   ENTER(T1820Log);
   if (0 == (ptr = VAPost(ptmlq, ptcb->pmsg, strcpy(BufNew(), "MLGetpML")))) {
      FATAL(NULL_PTR, "");
   }
   if (0 == (pml = *((pML *)ptr))) {
      FATAL(NULL_PTR, "");
   }
   BufDelete(ptr);
   if (0 == (pmlnode = GetMLNodeByName(pml, name))) {
      FATAL(NULL_PTR, "");
   }
   LOG = fopen(file_name, "a");
   mlread = pmlnode->pmldriver->mlread;
   for (;;) {
      // Wait until our device comes around.
      EventPend(pmlnode->psem, 0);
      if (0 > (mlread)(pml, pmlnode, (char *)0, 0, 0)) {
         if (0 > (ret = MLPathCheck(pml))) {
            VAPrintf(0, (VAPRINTF)vfprintf, pexec->LOG
             ,"%s: read failed: %s\n"
             ,ptcb->name
             ,ml_err[-ret]
            );
            EventSched(ptmlmsg, "1");
            TaskDelaySec(delay);
            continue;
         } 
         EventSched(ptmlmsg, "1");
         continue;
      }
      ptm = localtime((time_t *)&pmlnode->tv.tv_sec);
      VAPrintf(0, (VAPRINTF)vfprintf, LOG
       ,"%04d%02d%02d%02d%02d%02d %.3f\n"
       ,ptm->tm_year + 1900
       ,ptm->tm_mon + 1
       ,ptm->tm_mday
       ,ptm->tm_hour
       ,ptm->tm_min
       ,ptm->tm_sec
       ,pmlnode->dev.ds1820.temp / 65536.0
      );
      fflush(LOG);

      VAPrintf(0, (VAPRINTF)vfprintf, pexec->LOG
       ,"%04d%02d%02d%02d%02d%02d %8.3f %8.3f %s\n"
       ,ptm->tm_year + 1900
       ,ptm->tm_mon + 1
       ,ptm->tm_mday
       ,ptm->tm_hour
       ,ptm->tm_min
       ,ptm->tm_sec
       ,pmlnode->dev.ds1820.temp / 65536.0
       ,32.0 + ((1.8 / 65536.0) * pmlnode->dev.ds1820.temp)
       ,name
      );
      fflush(pexec->LOG);
      EventSched(ptmlmsg, "1");
      TaskDelaySec(delay);
   }
}
//=============================================================================

//=============================================================================
/*/
   A simple demo of a XidarML task object that reads or writes a DS2407.

   usage:

       spawn t2407 <task name> <bus queue> <bus mailbox> <log file> <log dir>

      e.g.:

       spawn t2407 siren_0  tmlq tmlmsg siren_0.log /tmp/foo/bar

   This task waits on his input queue for a command string, either

     read

   or
      write <channel> <state>   channel 0==A channel 1==B, state 1==conducting

      e.g.:

      write 1 0

   Both commands return the command status and then the switch status 
    byte in hexadecimal.
   
*/

int
T2407(
   char                         *cmd,
   int                           pri_ret
) {
   pTCB                          volatile ptcb;
   pEVENT                        volatile pq = 0;
   pEVENT                        volatile ptmlq = 0;
   pEVENT                        volatile ptmlmsg = 0;
   char                         *volatile file_name = 0;
   char                         *volatile name= 0;
   FILE                         *volatile LOG;

   int                           i;
   int                           j;
   int                           wr;
   char                         *ptr;
   char                         *rd_wr;
   char                         *channel;
   char                         *state;
   pML                           pml;
   pMLNODE                       pmlnode;
   struct tm                    *ptm;
   MLREAD                        mlread;
   MLWRITE                       mlwrite;
   MLONOFF                       mlon_off;
   va_list                       va_args;
   pEVENT                        pfrom;
   char                         *reply;

   if(0 == setjmp((ptcb = TaskCreate(&cmd, T2407_SIZ))->jmp_buf)) {
      ptcb->pmsg = EventCreate(EVENT_NAME_DEFAULT, EVENT_MSG);
      pq = QCreate(EVENT_NAME_DEFAULT, 32);
      SpawnArg(pri_ret, ptcb, ptr = Shift(&cmd), "null daemon queue name");
      SpawnArg(pri_ret, ptcb, ptmlq = GetEventByName(ptr), 
       "daemon queue not found");
      SpawnArg(pri_ret, ptcb, ptr = Shift(&cmd), "null daemon mailbox name");
      SpawnArg(pri_ret, ptcb, ptmlmsg = GetEventByName(ptr), 
       "daemon mailbox not found");
      SpawnArg(pri_ret, ptcb, ptr = Shift(&cmd), "null log file name");
      MALLOC(name, strlen(ptr) + 1, char);
      strcpy(name, ptr);
      SpawnArg(pri_ret, ptcb, ptr = Shift(&cmd), "null log file directory");
      MALLOC(file_name, strlen(ptr) + strlen(name) + 2, char);
      strcat(strcat(strcpy(file_name, ptr), "/"), name);
      SpawnAOK(pri_ret, ptcb);
   }
   StackClr(ptcb);
   ENTER(T2407);
   if (0 == (LOG = fopen(file_name, "a"))) {
      FATAL(NULL_PTR, ": fopen");
   }
   if (0 == (ptr = VAPost(ptmlq, ptcb->pmsg, strcpy(BufNew(), "MLGetpML")))) {
      FATAL(NULL_PTR, ": MLGetpML");
   }
   if (0 == (pml = *((pML *)ptr))) {
      FATAL(NULL_PTR, ": NULL pml");
   }
   BufDelete(ptr);
   if (0 == (pmlnode = GetMLNodeByName(pml, name))) {
      FATAL(NULL_PTR, ": GetMLNodeByName");
   }
   mlread = pmlnode->pmldriver->mlread;
   mlwrite = pmlnode->pmldriver->mlwrite;
   mlon_off = pmlnode->pmldriver->mlon_off;
   for (;;) {
      va_args = (va_list)EventPend(pq, 0);
      pfrom = va_arg(va_args, pEVENT);
      cmd = va_arg(va_args, char *);
      
      // Jump the queue.
      ptr = VAPost(ptmlq, ptcb->pmsg, 
       strcat(strcpy(BufNew(), "MLDoNode "), pmlnode->name));
      BufDelete(ptr);
      // Wait until our device comes around.
      EventPend(pmlnode->psem, 0);

      ptr = cmd;
      rd_wr = (rd_wr = Shift(&ptr)) ? rd_wr : "read";
      wr = !(strcasecmp(rd_wr, "write"));
      channel = (channel = Shift(&ptr)) ? channel : "0";
      state = (state = Shift(&ptr)) ? state : "0";
      if (wr) {
         (mlon_off)(pml, pmlnode, atoi(channel), !(atoi(state)));
      }
      (mlread)(pml, pmlnode, (char *)0, 0, 0);
      EventSched(ptmlmsg, "1");
      BufDelete(cmd);
      VAPrintf(0, (VAPRINTF)vsprintf, reply = BufNew()
       ,"1 %s %02x %02x"
       ,pmlnode->name
       ,pmlnode->dev.ds2407.channel_info_0
       ,pmlnode->dev.ds2407.channel_info_1
      );
      EventSched(pfrom, (void *)reply);
   }
}
//=============================================================================

