#include <stdio.h>
#include <assert.h>
#include <io.h>
#include <fcntl.h>
#include <wtypes.h>
#include <winbase.h>

class MLPORT;

class CONIN {
public:
   CONIN();
   virtual ~CONIN();
   char *Gets(char *buf);
   int Poll(int timeout);
private:
   HANDLE                   h_;
   HANDLE                   hthread_;
   DWORD                    thread_id_;
   HANDLE                   hevent_;
   BOOL                     ret_;
   DWORD                    cnt_;
   char                     buf_[256];
   friend DWORD WINAPI ThreadConin(LPVOID pconin);
};

CONIN::Poll(
   int                      timeout
) {
   int                      ret = 0;

   switch(WaitForSingleObject(hevent_, timeout)) {
      case WAIT_ABANDONED: {
         break;
      }
      case WAIT_OBJECT_0: {
         ret = cnt_;
         break;
      }
      case WAIT_TIMEOUT: {
         break;
      }
      case WAIT_FAILED: {
         printf("Waitfor WAIT_FAILED %d\n", GetLastError());
         break;
      }
      default: {
         assert(0);
         break;
      }
   }
   return(ret);
}

char *
CONIN::Gets(
   char                    *buf
) {
   if (cnt_) {
      strncpy(buf, buf_, cnt_);
      SetEvent(hevent_);
      return(buf);
   }
   return(0);
}

CONIN::CONIN(
) {
   hevent_ = CreateEvent(
    (LPSECURITY_ATTRIBUTES) 0
   ,(BOOL)                  FALSE  // Manual-reset.
   ,(BOOL)                  FALSE  // Initial state.
   ,(LPCTSTR)               0      // Name.
   );
   assert(0 != hevent_);

   hthread_ = CreateThread(
    (LPSECURITY_ATTRIBUTES)   0
   ,(DWORD)                   0
   ,(LPTHREAD_START_ROUTINE)  ThreadConin
   ,(LPVOID)                  this
   ,(DWORD)                   0
   ,(LPDWORD)                &thread_id_
   );
}

CONIN::~CONIN(
) {
    if (WaitForSingleObject(hthread_, 5000) == WAIT_TIMEOUT) {
        TerminateThread(hthread_, 1);
   }
}

DWORD WINAPI
ThreadConin(
   LPVOID                     p
) {
   CONIN                     *pconin;

   pconin = (CONIN *)p;

   pconin->h_ = CreateFile(
    (LPCTSTR)
      "CONIN$"
   ,(DWORD) 0
     | GENERIC_READ
   ,(DWORD) 0
     | FILE_SHARE_READ
     | FILE_SHARE_WRITE
   ,(LPSECURITY_ATTRIBUTES) 0
   ,(DWORD) 0
     | OPEN_EXISTING
   ,(DWORD) 0                    // Flags.
   ,(HANDLE) 0                   // Attributes.
   );
   assert(0 != pconin->h_);

   while (1) {
      pconin->cnt_ = 0;
      memset(pconin->buf_, 0, sizeof(pconin->buf_));
      pconin->ret_ = ReadFile(
       pconin->h_, pconin->buf_, 255, &pconin->cnt_, (LPOVERLAPPED) 0);
      SetEvent(pconin->hevent_);
      WaitForSingleObject(pconin->hevent_, INFINITE);
   }
}

#define dcbpr(dcbmem) printf("%04x " #dcbmem "\n", dcbmem)
void
DCBPr(
   DCB            *pdcb
) {
   dcbpr(pdcb->DCBlength);
   dcbpr(pdcb->BaudRate);
   dcbpr(pdcb->fBinary);
   dcbpr(pdcb->fParity);
   dcbpr(pdcb->fOutxCtsFlow);
   dcbpr(pdcb->fOutxDsrFlow);
   dcbpr(pdcb->fDtrControl);
   dcbpr(pdcb->fDsrSensitivity);
   dcbpr(pdcb->fTXContinueOnXoff);

   dcbpr(pdcb->fOutX);
   dcbpr(pdcb->fInX);
   dcbpr(pdcb->fErrorChar);
   dcbpr(pdcb->fNull);
   dcbpr(pdcb->fRtsControl);
   dcbpr(pdcb->fAbortOnError);
   dcbpr(pdcb->fDummy2);
   dcbpr(pdcb->wReserved);
   dcbpr(pdcb->XonLim);

   dcbpr(pdcb->XoffLim);
   dcbpr(pdcb->ByteSize);
   dcbpr(pdcb->Parity);
   dcbpr(pdcb->StopBits);
   dcbpr(pdcb->XonChar);
   dcbpr(pdcb->XoffChar);
   dcbpr(pdcb->ErrorChar);
   dcbpr(pdcb->EofChar);
   dcbpr(pdcb->EvtChar);

   dcbpr(pdcb->wReserved1);
}

class MLPORT {
public:
   MLPORT();
   virtual ~MLPORT();
   int MLWrite(char *buf, int cnt, int flag);
   int MLOpen(char *name);
private:
   HANDLE                   h_;
   COMMTIMEOUTS             commtimeouts_;
   DCB                      dcb_;
};

MLPORT::MLPORT(
) {
}
MLPORT::~MLPORT(
) {
}

int
MLPORT::MLOpen(
   char                    *name
){
   h_ = CreateFile(
    name
   ,GENERIC_READ | GENERIC_WRITE
   ,0
   ,NULL
   ,OPEN_EXISTING
   ,FILE_ATTRIBUTE_NORMAL
   ,0
   );
   commtimeouts_.ReadIntervalTimeout = 20;
   commtimeouts_.ReadTotalTimeoutMultiplier = 1;
   commtimeouts_.ReadTotalTimeoutConstant =  10;
   commtimeouts_.WriteTotalTimeoutMultiplier = 0;
   commtimeouts_.WriteTotalTimeoutConstant = 0;
   SetCommTimeouts(h_, &commtimeouts_);

   GetCommState(h_, &dcb_);
// DCBPr(&dcb_);
   dcb_.DCBlength = sizeof(DCB);               // sizeof(DCB)
   dcb_.BaudRate = 115200;                     // current baud rate
   dcb_.fBinary = 1;                           // binary mode, no EOF check
   dcb_.fParity = 0;                           // enable parity checking
   dcb_.fOutxCtsFlow = 1;                      // CTS output flow control
   dcb_.fOutxDsrFlow = 1;                      // DSR output flow control
   dcb_.fDtrControl = DTR_CONTROL_ENABLE;      // DTR flow control type
   dcb_.fDsrSensitivity = 1;                   // DSR sensitivity
   dcb_.fTXContinueOnXoff = 0;                 // XOFF continues Tx
   dcb_.fOutX = 0;                             // XON/XOFF out flow control
   dcb_.fInX = 0;                              // XON/XOFF in flow control
   dcb_.fErrorChar = 0;                        // enable error replacement
   dcb_.fNull = 0;                             // enable null stripping
   dcb_.fRtsControl = RTS_CONTROL_ENABLE;      // RTS flow control
   dcb_.fAbortOnError = 0;                     // abort reads/writes on error
   dcb_.fDummy2 = 0;                           // reserved
   dcb_.wReserved = 0;                         // not currently used
   dcb_.XonLim = -1;                           // transmit XON threshold
   dcb_.XoffLim = -1;                          // transmit XOFF threshold
   dcb_.ByteSize = 8;                          // number of bits/byte, 4-8
   dcb_.Parity = NOPARITY;                     // 0-4=no,odd,even,mark,space
   dcb_.StopBits = ONESTOPBIT;                 // 0,1,2 = 1, 1.5, 2
   dcb_.XonChar = 0;                           // Tx and Rx XON character
   dcb_.XoffChar = 0;                          // Tx and Rx XOFF character
   dcb_.ErrorChar = 0;                         // error replacement character
   dcb_.EofChar = 0;                           // end of input character
   dcb_.EvtChar = 0;                           // received event character
   dcb_.wReserved1 = 0;                        // reserved; do not use
   SetCommState(h_, &dcb_);
}

int
MLPORT::MLWrite(
   char                    *buf,
   int                      cnt_req,
   int                      flag
) {
   BOOL                     ret;
   DWORD                    cnt_wr = 0;
   DWORD                    cnt_rd = 0;
   char                     cuf[256];

// PurgeComm(h_, PURGE_TXCLEAR | PURGE_RXCLEAR);
   ret = WriteFile(h_, buf, cnt_req,  &cnt_wr, NULL);
printf("ret %d cnt wr %d\n", ret, cnt_wr);
   ret = ReadFile(h_, cuf, cnt_req,  &cnt_rd, NULL);
printf("ret %d cnt rd %d\n", ret, cnt_rd);
   memcpy(buf, cuf, cnt_req);
   return(cnt_rd);
}

int
main(
   void
) {
   CONIN                    conin;
   MLPORT                   mlport;
   char                     buf[256];

	mlport.MLOpen("COM2");

   while (1) {
      if (conin.Poll(1000)) {
         printf("conin.Gets %s", conin.Gets(buf));
      } else {
         puts("timed out");
         strcpy(buf, "Come in Rangoon, this is Taipei.  Do you read me?");
         mlport.MLWrite(buf, strlen(buf) + 1, 0);
         puts(buf);
      }
   }
}
